""" Loop unrolling transformation in ufl

NB: Doing this in UFL is not the best level of abstractions to do this.
    However, for the sum factorization code paths, there are some advantages
    in traversing unrolled expression trees.
"""

from ufl.algorithms import MultiFunction
from dune.perftool.ufl.transformations import ufl_transformation
from dune.perftool.ufl.flatoperators import construct_binary_operator
from dune.perftool.options import get_option

import ufl.classes as uc


class UnrollDimensionLoops(MultiFunction):
    def __init__(self):
        self.replace = {}
        MultiFunction.__init__(self)

    call = MultiFunction.__call__

    def expr(self, o):
        return self.reuse_if_untouched(o, *tuple(self(op) for op in o.ufl_operands))

    def multi_index(self, o):
        return uc.MultiIndex(tuple(self.replace.get(i, i) for i in o))

    def index_sum(self, o):
        operands = []

        # TODO: What is the correct way to get the shape here?
        for i in range(o.geometric_dimension()):
            self.replace[o.ufl_operands[1][0]] = uc.FixedIndex(i)
            operands.append(self.call(o.ufl_operands[0]))
            del self.replace[o.ufl_operands[1][0]]

        return construct_binary_operator(tuple(operands), uc.Sum)


@ufl_transformation(name="unroll")
def unroll_dimension_loops(expr):
    if get_option("unroll_dimension_loops"):
        return UnrollDimensionLoops()(expr)
    else:
        return expr
